import typing


def curried_binary_search_iterative(
    predicate: typing.Callable[[int], bool],
    decorate_with: typing.Callable[[typing.Callable], typing.Callable] = lambda x: x,
) -> typing.Callable[[int, int], typing.Optional[int]]:
    """Find the positive integer threshold below which a search criteria is
    never satisfied and above which it is always satisfied.

    Currying allows for calls that don't pass the predicate callable
    as an argument, which is necessary for compatibility with Numba jit
    compilation.

    Parameters
    ----------
    predicate : callable object
        Returns whether an integer value satisfies the search criteria.
    decorate_with : callable, optional
        A decorator that will be applied to the inner search function. Useful
        to enable jit compilation.

    Returns
    -------
    searcher : Callable[[int, int], Optional[int]]
        The curried search function, which takes an optional integer
        `lower_bound` and `upper_bound` and returns the first integer value
        that satisfies the search criteria.

        If no value satisfies the search criteria, the search function
        will return None.
    """

    @decorate_with
    def binary_search_iterative(
        lower_bound: int,
        upper_bound: int,
    ) -> typing.Optional[int]:
        """Find the positive integer threshold below which a search criteria is
        never satisfied and above which it is always satisfied.

        Parameters
        ----------
        lower_bound : int
            Lower bound for the binary search, inclusive.
        upper_bound : int
            Upper bound for the binary search, inclusive.

        Returns
        -------
        guess
            The lowest integer value that satisfies the search criteria, and
            None if upper_bound does not satisfy the search criteria or search
            range is empty (i.e., lower_bound > upper_bound).
        """

        # Check if the range is invalid.
        if lower_bound > upper_bound:
            return None

        # Loop until lower_bound is greater than upper_bound
        while lower_bound <= upper_bound:
            # Find the midpoint
            midpoint = (lower_bound + upper_bound) // 2

            # If predicate is satisfied by midpoint, search the left half.
            if predicate(midpoint):
                upper_bound = midpoint - 1
            # Otherwise, search the right half.
            else:
                lower_bound = midpoint + 1

        # If the search ended successfully, lower_bound is the answer.
        # Otherwise, return None.
        if predicate(lower_bound):
            return lower_bound
        else:
            return None

    return binary_search_iterative



#!/usr/bin/env python

'''
`doubling_search` tests for `interval_search` package.
'''

import interval_search as inch


def test_doubling_search_trivial():
    assert inch.doubling_search(lambda __: True) == 0
    assert inch.doubling_search(lambda __: True, 10) == 10


def test_doubling_search_nontrivial():
    assert inch.doubling_search(lambda x: x >= 5) == 5
    assert inch.doubling_search(lambda x: x >= 5, 10) == 10
    assert inch.doubling_search(lambda x: x >= 422) == 422
    assert inch.doubling_search(lambda x: x >= 422, 10) == 422
    assert inch.doubling_search(lambda x: x >= 423) == 423
    assert inch.doubling_search(lambda x: x >= 423, 10) == 423

def test_doubling_search_nontrivial_satisfaction_bound():
    assert inch.doubling_search(lambda x: x >= 5, satisfaction_bound = 3) == 3
    assert inch.doubling_search(lambda x: x >= 5, satisfaction_bound = 5) == 5
    assert inch.doubling_search(lambda x: x >= 5, satisfaction_bound = 6) == 5
    assert inch.doubling_search(lambda x: x >= 5, 10, 0) == 10
    assert inch.doubling_search(lambda x: x >= 50, 10, 0) == 10
    assert inch.doubling_search(lambda x: x >= 422, 10, 10) == 10
    assert inch.doubling_search(lambda x: x >= 422, 10, 422) == 422
    assert inch.doubling_search(lambda x: x >= 422, 10, 430) == 422



#!/usr/bin/env python

'''
`curried_binary_search` tests for `interval_search` package.
'''

import interval_search as inch


def test_curried_binary_search_singleton():
    assert inch.curried_binary_search(lambda __: True)(10, 10) == 10


def test_curried_binary_search():
    assert inch.curried_binary_search(lambda x: x >= 5)(0, 100) == 5


def test_fruitless_curried_binary_search():
    assert inch.curried_binary_search(lambda x: False)(0, 100) is None


def test_empty_curried_binary_search():
    assert inch.curried_binary_search(lambda x: False)(100, 99) is None
    assert inch.curried_binary_search(lambda x: False)(100, 0) is None
    assert inch.curried_binary_search(lambda x: True)(100, 99) is None
    assert inch.curried_binary_search(lambda x: True)(100, 0) is None



import typing

from .curried_doubling_search import curried_doubling_search


def doubling_search(
    predicate: typing.Callable[[int], bool],
    lower_bound: int = 0,
    satisfaction_bound: typing.Optional[int] = None,
) -> int:
    """Find the positive integer threshold below which a search criteria is
    never satisfied and above which it is always satisfied.

    Parameters
    ----------
    predicate : callable object
        Returns whether an integer value satisfies the search criteria.
    lower_bound : int, optional
        The initial guess. Should be less than or equal to the first value that
        satsfies the search criteria. Used for recursion. Default is 0.
    satisfaction_bound : int, default None
        Any integer values at or past this value are considered to satisfy the
        search criteria without calling predicate.

    Returns
    -------
    threshold
        The lowest integer value that satisfies the search criteria.
    """
    if satisfaction_bound is not None:
        predicate_ = lambda x: (x >= satisfaction_bound) or predicate(x)
    else:
        predicate_ = predicate
    return curried_doubling_search(predicate_)(lower_bound)



import typing

from .curried_binary_search_recursive import curried_binary_search_recursive

def binary_search_recursive(
    predicate: typing.Callable[[int], bool],
    lower_bound: int,
    upper_bound: int,
) -> typing.Optional[int]:
    """Find the positive integer threshold below which a search criteria is
    never satisfied and above which it is always satisfied.

    Parameters
    ----------
    predicate : callable object
        Returns whether an integer value satisfies the search criteria.
    lower_bound : int
        Lower bound for the binary search, inclusive.
    upper_bound : int
        Upper bound for the binary search, inclusive.

    Returns
    -------
    guess
        The lowest integer value that satisfies the search criteria, and None
        if upper_bound does not satisfy the search criteria or search range is
        empty (i.e., lower_bound > upper_bound).
    """

    return curried_binary_search_recursive(predicate)(lower_bound, upper_bound)



#!/usr/bin/env python

"""
`curried_binary_search_iterative` tests for `interval_search` package.
"""

import interval_search as inch


def test_curried_binary_search_iterative_singleton():
    assert inch.curried_binary_search_iterative(lambda __: True)(10, 10) == 10


def test_curried_binary_search_iterative():
    assert inch.curried_binary_search_iterative(lambda x: x >= 5)(0, 100) == 5


def test_fruitless_curried_binary_search_iterative():
    assert inch.curried_binary_search_iterative(lambda x: False)(
        0, 100
    ) is None


def test_empty_curried_binary_search_iterative():
    assert inch.curried_binary_search_iterative(lambda x: False)(
        100, 99
    ) is None
    assert inch.curried_binary_search_iterative(lambda x: False)(
        100, 0
    ) is None
    assert inch.curried_binary_search_iterative(lambda x: True)(
        100, 99
    ) is None
    assert inch.curried_binary_search_iterative(lambda x: True)(100, 0) is None



import typing

from .binary_search_iterative import binary_search_iterative


def curried_doubling_search(
    predicate: typing.Callable[[int], bool],
    decorate_with: typing.Callable[
        [typing.Callable], typing.Callable
    ] = lambda x: x,
) -> typing.Callable[[typing.Optional[int]], int]:
    """Find the positive integer threshold below which a search criteria is
    never satisfied and above which it is always satisfied.

    Currying allows for recursive calls that don't pass the predicate callable
    as an argument, which is necessary for compatibility with Numba jit
    compilation.

    Parameters
    ----------
    predicate : callable object
        Returns whether an integer value satisfies the search criteria.
    decorate_with : callable, optional
        A decorator that will be applied to the inner search function. Useful
        to enable jit compilation.

    Returns
    -------
    searcher : Callable[[Optional[int]], int]
        The curried search function, which takes an optional integer
        `lower_bound` and returns an integer value that satisfies the search
        criteria. If no lower bound is provided, the default is taken as 0.
    """

    @decorate_with
    def doubling_search(lower_bound: int = 0) -> int:
        """Find the positive integer threshold below which a search criteria is
        never satisfied and above which it is always satisfied.

        Parameters
        ----------
        lower_bound : int, optional
            The initial guess. Should be less than or equal to the first value that
            satsfies the search criteria. Used for recursion. Default is 0.

        Returns
        -------
        threshold : int
            The lowest integer value that satisfies the search criteria.
        """

        assert lower_bound >= 0, lower_bound

        bound = 1
        while not predicate(lower_bound + bound):
            bound *= 2

        prev_bound = bound // 2
        prev_guess = lower_bound + prev_bound
        return binary_search_iterative(
            predicate, prev_guess, lower_bound + bound
        )

    return doubling_search



import typing


def curried_binary_search_recursive(
    predicate: typing.Callable[[int], bool],
    decorate_with: typing.Callable[
        [typing.Callable], typing.Callable
    ] = lambda x: x,
) -> typing.Callable[[int, int], typing.Optional[int]]:
    """Find the positive integer threshold below which a search criteria is
    never satisfied and above which it is always satisfied.

    Currying allows for recursive calls that don't pass the predicate callable
    as an argument, which is necessary for compatibility with Numba jit
    compilation.

    Parameters
    ----------
    predicate : callable object
        Returns whether an integer value satisfies the search criteria.
    decorate_with : callable, optional
        A decorator that will be applied to the inner search function. Useful
        to enable jit compilation.

    Returns
    -------
    searcher : Callable[[int, int], Optional[int]]
        The curried search function, which takes an optional integer
        `lower_bound` and `upper_bound` and returns the first integer value
        that satisfies the search criteria.

        If no value satisfies the search criteria, the curried search function
        will return None.
    """

    @decorate_with
    def binary_search_recursive(
        lower_bound: int,
        upper_bound: int,
    ) -> typing.Optional[int]:
        """Find the positive integer threshold below which a search criteria is
        never satisfied and above which it is always satisfied.

        Parameters
        ----------
        lower_bound : int
            Lower bound for the binary search, inclusive.
        upper_bound : int
            Upper bound for the binary search, inclusive.

        Returns
        -------
        guess
            The lowest integer value that satisfies the search criteria, and
            None if upper_bound does not satisfy the search criteria or search
            range is empty (i.e., lower_bound > upper_bound).
        """

        if lower_bound > upper_bound:
            return None
        if lower_bound == upper_bound:
            if predicate(lower_bound):
                return lower_bound
            else:
                return None

        midpoint = (lower_bound + upper_bound) // 2

        if predicate(midpoint):
            return binary_search_recursive(lower_bound, midpoint)
        else:
            return binary_search_recursive(midpoint + 1, upper_bound)

    return binary_search_recursive



#!/usr/bin/env python

'''
`curried_doubling_search` tests for `interval_search` package.
'''

import interval_search as inch


def test_curried_doubling_search_trivial():
    assert inch.curried_doubling_search(lambda __: True)() == 0
    assert inch.curried_doubling_search(lambda __: True)(10) == 10


def test_curried_doubling_search_nontrivial():
    assert inch.curried_doubling_search(lambda x: x >= 5)() == 5
    assert inch.curried_doubling_search(lambda x: x >= 5)(10) == 10
    assert inch.curried_doubling_search(lambda x: x >= 422)() == 422
    assert inch.curried_doubling_search(lambda x: x >= 422)(10) == 422
    assert inch.curried_doubling_search(lambda x: x >= 423)() == 423
    assert inch.curried_doubling_search(lambda x: x >= 423)(10) == 423
